# Copyright (C) 2018 Lukas Vrabec, <lvrabec@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

cap = {
        'CHOWN':'chown ',
        'DAC_OVERRIDE':'dac_override ',
        'DAC_READ_SEARCH':'dac_read_search ',
        'FOWNER':'fowner ',
        'FSETID':'fsetid ',
        'KILL':'kill ',
        'SETGID':'setgid ',
        'SETUID':'setuid ',
        'SETPCAP':'setpcap ',
        'LINUX_IMMUTABLE':'linux_immutable ',
        'NET_BIND_SERVICE':'net_bind_service ',
        'NET_BROADCAST':'net_broadcast ',
        'NET_ADMIN':'net_admin ',
        'NET_RAW':'net_raw ',
        'IPC_LOCK':'ipc_lock ',
        'IPC_OWNER':'ipc_owner ',
        'SYS_MODULE':'sys_module ',
        'SYS_RAWIO':'sys_rawio ',
        'SYS_CHROOT':'sys_chroot ',
        'SYS_PTRACE':'sys_ptrace ',
        'SYS_PACCT':'sys_pacct ',
        'SYS_ADMIN':'sys_admin ',
        'SYS_BOOT':'sys_boot ',
        'SYS_NICE':'sys_nice ',
        'SYS_RESOURCE':'sys_resource ',
        'SYS_TIME':'sys_time ',
        'SYS_TTY_CONFIG':'sys_tty_config ',
        'MKNOD':'mknod ',
        'LEASE':'lease ',
        'AUDIT_WRITE':'audit_write ',
        'AUDIT_CONTROL':'audit_control ',
        'SETFCAP':'setfcap ',
        'MAC_OVERRIDE':'mac_override ',
        'MAC_ADMIN':'mac_admin ',
        'SYSLOG':'syslog ',
        'WAKE_ALARM':'wake_alarm ',
        'BLOCK_SUSPEND':'block_suspend ',
        'AUDIT_READ':'audit_read '
}

perm = {
        'drw': 'open read getattr lock search ioctl add_name remove_name write',
        'dro': 'getattr search open read lock ioctl',
        'frw': 'getattr read write append ioctl lock map open create ',
        'fro': 'getattr read ioctl lock open ',
        'srw': 'getattr read write append open ',
        'sro': 'getattr read open ',
}

socket = {
        'tcp': 'tcp_socket',
        'udp': 'udp_socket',
}



# Copyright (C) 2018 Lukas Vrabec, <lvrabec@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json

def parse_inspect(data):
    json_rep = json.loads(data)
    if 'container=podman' not in json_rep[0]['Config']['Env']:
        for item in json_rep[0]['Mounts']:
            item['source'] = item['Source']
            if item['Mode'] == 'rw':
                item['options'] = 'rw'
            if item['Mode'] == 'ro':
                item['options'] = 'ro'

        temp_ports = []

        for item in json_rep[0]['NetworkSettings']['Ports']:
            container_port = item.split('/')
            host_port = json_rep[0]['NetworkSettings']['Ports'][item][0]['HostPort']
            new_port = {'hostPort':int(host_port), 'protocol':container_port[1]}
            temp_ports.append(new_port)

        del json_rep[0]['NetworkSettings']['Ports']

        json_rep[0]['NetworkSettings']['Ports'] = temp_ports

    return json_rep

def parse_cap(data):
    return data.decode().split('\n')[1].split(',')

def parse_is_podman(data):
    json_rep = json.loads(data)
    return 'container=podman' in json_rep[0]['Config']['Env']

def context_to_type(context):
    return context.split('=')[1].split(':')[2]

def remove_dupe_perms(string):
    perms = string.split()
    return " ".join(sorted(set(perms), key=perms.index))

def parse_avc_file(data):
    append_rules = []

    for avc in data.splitlines():
        new_rule = []
        items = avc.split(' ')

        if 'type=AVC' not in items[0]:
            continue

        for item in items:
            if 'scontext' in item:
                new_rule.append(context_to_type(item))
            if 'tcontext' in item:
                new_rule.append(context_to_type(item))
            if 'tclass' in item:
                new_rule.append(item.split('=')[1])

        open_bracket = items.index('{')
        perm = open_bracket+1
        new_rule.append(items[perm])

        for rule in append_rules:
            if rule[0] == new_rule[0] and rule[1] == new_rule[1] and rule[2] == new_rule[2]:
                new_rule[3] = rule[3] + ' ' + new_rule[3]
                append_rules.remove(rule)
        new_rule[3] = remove_dupe_perms(new_rule[3])
        append_rules.append(new_rule)

    return append_rules



# Copyright (C) 2018 Lukas Vrabec, <lvrabec@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import selinux
import semanage

from shutil import copy
from os import chdir, getcwd, write, read, remove, replace
from os.path import exists

import tarfile

import udica.perms as perms

# Check if templates are present in current directory
# if yes, the templates directory is used instead of system one.
TEMPLATES_STORE = './templates' if exists('./templates') else '/usr/share/udica/templates'

CONFIG_CONTAINER = '/etc'
HOME_CONTAINER = '/home'
LOG_CONTAINER = '/var/log'
TMP_CONTAINER = '/tmp'

TEMPLATE_PLAYBOOK = '/usr/share/udica/ansible/deploy-module.yml'
VARIABLE_FILE_NAME = 'variables-deploy-module.yml'

templates_to_load = []

def add_template(template):
    templates_to_load.append(template)

def list_templates_to_string(templates_to_load):
    return '.cil,'.join(map(str, templates_to_load)) + '.cil'

def list_contexts(directory):
    directory_len = (len(directory))

    handle = semanage.semanage_handle_create()
    semanage.semanage_connect(handle)

    (rc, fclist) = semanage.semanage_fcontext_list(handle)
    (rc, fclocal) = semanage.semanage_fcontext_list_local(handle)
    (rc, fchome) = semanage.semanage_fcontext_list_homedirs(handle)

    contexts = []
    for fcontext in fclist + fclocal + fchome:
        expression = semanage.semanage_fcontext_get_expr(fcontext)
        if expression[0:directory_len] == directory:
            context = semanage.semanage_fcontext_get_con(fcontext)
            if context:
                contexts.append(semanage.semanage_context_get_type(context))

    selabel = selinux.selabel_open(selinux.SELABEL_CTX_FILE, None, 0)
    (rc, context) = selinux.selabel_lookup(selabel, directory, 0)
    if context == None:
        if exists(directory) == False:
            exit(1)
        context = selinux.getfilecon(directory)[1]
    contexts.append(context.split(':')[2])
    return contexts

def list_ports(port_number, port_proto):

    handle = semanage.semanage_handle_create()
    semanage.semanage_connect(handle)

    (rc, plist) = semanage.semanage_port_list(handle)
    (rc, plocal) = semanage.semanage_port_list_local(handle)

    for port in plocal + plist:
        con = semanage.semanage_port_get_con(port)
        ctype = semanage.semanage_context_get_type(con)
        proto = semanage.semanage_port_get_proto(port)
        proto_str = semanage.semanage_port_get_proto_str(proto)
        low = semanage.semanage_port_get_low(port)
        high = semanage.semanage_port_get_high(port)
        if low <= port_number <= high and port_proto == proto_str:
            return ctype

def create_policy(opts, capabilities, mounts, ports, append_rules):
    policy = open(opts['ContainerName'] +'.cil', 'w')
    policy.write('(block ' + opts['ContainerName'] + '\n')
    policy.write('    (blockinherit container)\n')
    add_template("base_container")

    if opts['FullNetworkAccess']:
        policy.write('    (blockinherit net_container)\n')
        add_template("net_container")

    if opts['VirtAccess']:
        policy.write('    (blockinherit virt_container)\n')
        add_template("virt_container")

    if opts['XAccess']:
        policy.write('    (blockinherit x_container)\n')
        add_template("x_container")

    if opts['TtyAccess']:
        policy.write('    (blockinherit tty_container)\n')
        add_template("tty_container")

    if ports:
        policy.write('    (blockinherit restricted_net_container)\n')
        add_template("net_container")

    if opts['StreamConnect']:
        policy.write('    (allow process ' + opts['StreamConnect'] + '.process ( unix_stream_socket ( connectto ))) \n')
        policy.write('    (allow process ' + opts['StreamConnect'] + '.socket ( sock_file ( getattr write open append ))) \n')

    # capabilities
    if capabilities:
        caps = ''
        for item in capabilities:
            # Capabilities parsed from podman inspection JSON file have prefix "CAP_", this should be removed
            if "CAP_" in item:
                caps = caps + perms.cap[item[4:]]
            else:
                caps = caps + perms.cap[item]

        policy.write('    (allow process process ( capability ( ' + caps  + '))) \n')
        policy.write('\n')

    # ports
    for item in ports:
        if 'hostPort' in item:
            policy.write('    (allow process ' + list_ports(item['hostPort'], item['protocol']) + ' ( ' + perms.socket[item['protocol']] + ' (  name_bind ))) \n')

    # mounts
    for item in mounts:
        if not item['Source'].find("/"):
            if (item['Source'] == LOG_CONTAINER and item['RW'] is False):
                policy.write('    (blockinherit log_container)\n')
                add_template("log_container")
                continue

            if (item['Source'] == LOG_CONTAINER and item['RW'] is True):
                policy.write('    (blockinherit log_rw_container)\n')
                add_template("log_container")
                continue

            if (item['Source'] == HOME_CONTAINER and item['RW'] is False):
                policy.write('    (blockinherit home_container)\n')
                add_template("home_container")
                continue

            if (item['Source'] == HOME_CONTAINER and item['RW'] is True):
                policy.write('    (blockinherit home_rw_container)\n')
                add_template("home_container")
                continue

            if (item['Source'] == TMP_CONTAINER and item['RW'] is False):
                policy.write('    (blockinherit tmp_container)\n')
                add_template("tmp_container")
                continue

            if (item['Source'] == TMP_CONTAINER and item['RW'] is True):
                policy.write('    (blockinherit tmp_rw_container)\n')
                add_template("tmp_container")
                continue

            if (item['Source'] == CONFIG_CONTAINER and item['RW'] is False):
                policy.write('    (blockinherit config_container)\n')
                add_template("config_container")
                continue

            if (item['Source'] == CONFIG_CONTAINER and item['RW'] is True):
                policy.write('    (blockinherit config_rw_container)\n')
                add_template("config_container")
                continue

            contexts = list_contexts(item['Source'])
            for context in contexts:
                if item['RW'] is True:
                    policy.write('    (allow process ' + context + ' ( dir ( ' + perms.perm['drw'] + ' ))) \n')
                    policy.write('    (allow process ' + context + ' ( file ( ' + perms.perm['frw'] + ' ))) \n')
                    policy.write('    (allow process ' + context + ' ( sock_file ( ' + perms.perm['srw'] + ' ))) \n')
                if item['RW'] is False:
                    policy.write('    (allow process ' + context + ' ( dir ( ' + perms.perm['dro'] + ' ))) \n')
                    policy.write('    (allow process ' + context + ' ( file ( ' + perms.perm['fro'] + ' ))) \n')
                    policy.write('    (allow process ' + context + ' ( sock_file ( ' + perms.perm['sro'] + ' ))) \n')

    if append_rules != None:
        for rule in append_rules:
            if opts['ContainerName'] in rule[0]:
                policy.write('    (allow process ' + rule[1] + ' ( ' + rule[2] + ' ( ' + rule[3] + ' ))) \n')
            else:
                print('WARNING: process type: ' + rule[0] + ' seems to be unrelated to this container policy. Skipping allow rule.')

    policy.write(')')
    policy.close()

def load_policy(opts):
    PWD = getcwd()
    chdir(TEMPLATES_STORE)

    if opts['LoadModules']:
        handle = semanage.semanage_handle_create()
        semanage.semanage_connect(handle)

        for template in templates_to_load:
            semanage.semanage_module_install_file(handle, template + '.cil')

        chdir(PWD)

        semanage.semanage_module_install_file(handle, opts['ContainerName'] + '.cil')

        semanage.semanage_commit(handle)
    else:
        templates = list_templates_to_string(templates_to_load)
        if len(templates_to_load) > 1:
            print('\nPlease load these modules using: \n# semodule -i ' + opts['ContainerName'] + '.cil ' + TEMPLATES_STORE + "/{" + templates + '}')
        else:
            print('\nPlease load these modules using: \n# semodule -i ' + opts['ContainerName'] + '.cil ' + TEMPLATES_STORE + "/" + templates + '')

        chdir(PWD)

def generate_playbook(opts):
    src = TEMPLATE_PLAYBOOK
    dst = "./"
    copy(src, dst)

    with open(VARIABLE_FILE_NAME, 'w') as varsfile:
        varsfile.write('archive: ' + opts['ContainerName'] + '-policy.tar.gz\n')
        varsfile.write('policy: ' + opts['ContainerName'] + '.cil ' +
                       list_templates_to_string(templates_to_load).replace(',', ' ') + '\n')
        varsfile.write('final_policy: ' + opts['ContainerName'] + '.cil')

    with tarfile.open(opts['ContainerName'] + '-policy.tar.gz', 'w:gz') as tar:
        for template in templates_to_load:
            tar.add(TEMPLATES_STORE + '/' + template + '.cil', template + '.cil')
        tar.add(opts['ContainerName'] + '.cil')
        remove(opts['ContainerName'] +'.cil')

    print('\nAnsible playbook and archive with udica policies generated! \n' +
          'Please run ansible play to deploy the policy.')



# Copyright (C) 2018 Lukas Vrabec, <lvrabec@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import subprocess
import argparse
import shutil

# import udica
from udica.parse import parse_inspect, parse_cap, parse_is_podman, parse_avc_file
from udica.policy import create_policy, load_policy, generate_playbook

def get_args():
    parser = argparse.ArgumentParser(description='Script generates SELinux policy for running container.')
    parser.add_argument(
        type=str, help='Name for SELinux policy module', dest='ContainerName')
    parser.add_argument(
        '-i', '--container-id', type=str, help='Running container ID', dest='ContainerID', default=None)
    parser.add_argument(
        '-j', '--json', help='Load json from this file, use "-j -" for stdin', required=False, dest='JsonFile', default=None)
    parser.add_argument(
        '--full-network-access', help='Allow container full Network access ', required=False, dest='FullNetworkAccess', action='store_true')
    parser.add_argument(
        '--tty-access', help='Allow container to read and write the controlling terminal ', required=False, dest='TtyAccess', action='store_true')
    parser.add_argument(
        '--X-access', help='Allow container to communicate with Xserver ', required=False, dest='XAccess', action='store_true')
    parser.add_argument(
        '--virt-access', help='Allow container to communicate with libvirt ', required=False, dest='VirtAccess', action='store_true')
    parser.add_argument(
        '-s','--stream-connect', help='Allow container to stream connect with given SELinux domain ', required=False, dest='StreamConnect')
    parser.add_argument(
        '-l', '--load-modules', help='Load templates and module created by this tool ', required=False, dest='LoadModules', action='store_true')
    parser.add_argument(
        '-c', '--caps', help='List of capabilities, e.g "-c AUDIT_WRITE,CHOWN,DAC_OVERRIDE,FOWNER,FSETID,KILL,MKNOD,NET_BIND_SERVICE,NET_RAW,SETFCAP,SETGID,SETPCAP,SETUID,SYS_CHROOT"', required=False, dest='Caps', default=None)
    parser.add_argument(
        '-d', '--ansible', help='Generate ansible playbook to deploy SELinux policy for containers ', required=False, dest='Ansible', action='store_true')
    parser.add_argument(
        '-a', '--append-rules', type=str, help='Append more SELinux allow rules from file', dest='FileAVCS', required=False, default=None)
    args = parser.parse_args()
    return vars(args)

def main():

    opts = get_args()

    if opts['ContainerID']:
        container_inspect_data = None
        for backend in ["podman", "docker"]:
            try:
                run_inspect = subprocess.Popen([backend, "inspect", opts['ContainerID']], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
                inspect_data = run_inspect.communicate()[0]
                if run_inspect.returncode != 0:
                    inspect_data = None
            except FileNotFoundError:
                inspect_data = None

            if inspect_data:
                container_inspect_data = inspect_data
                break

        if not container_inspect_data:
            print('Container with specified ID does not exits!')
            exit(3)

    if opts['JsonFile']:
        if opts['JsonFile'] == '-':
            import sys
            container_inspect_data = sys.stdin.read()
        else:
            import os.path
            if os.path.isfile(opts['JsonFile']):
                with open(opts['JsonFile'], 'r') as f:
                    container_inspect_data = f.read()
            else:
                print('Json file does not exists!')
                exit(3)

    if (not opts['JsonFile']) and (not opts['ContainerID']):
        try:
            import sys
            container_inspect_data = sys.stdin.read()
        except Exception as e:
            print('Couldn\'t parse inspect data from stdin:', e)
            exit(3)

    try:
        container_inspect = parse_inspect(container_inspect_data)
    except Exception as e:
        print('Couldn\'t parse inspect data:', e)
        exit(3)
    container_mounts = container_inspect[0]['Mounts']
    container_ports = container_inspect[0]['NetworkSettings']['Ports']

    try:
        is_podman = parse_is_podman(container_inspect_data)
    except Exception as e:
        print('Couldn\'t parse podman:', e)
        exit(3)

    # Append allow rules if AVCs log is provided
    append_rules = None
    if opts['FileAVCS']:
        import os.path
        if os.path.isfile(opts['FileAVCS']):
            with open(opts['FileAVCS'], 'r') as f:
                try:
                    append_rules = parse_avc_file(f.read())
                except Exception as e:
                    print('Couldn\'t parse inspect data:', e)
                    exit(3)
            f.close()
        else:
            print('AVC file does not exists!')
            exit(3)

    container_caps = []

    if opts['Caps']:
        if opts['Caps'] == 'None':
            container_caps = []
        else:
            container_caps = opts['Caps'].split(',')
    else:
        if is_podman:
            container_caps = container_inspect[0]['EffectiveCaps']

    try:
        create_policy(opts, container_caps, container_mounts, container_ports, append_rules)
    except Exception as e:
        print('Couldn\'t create policy:', e)
        exit(4)

    print('\nPolicy ' + opts['ContainerName'] + ' created!')

    if opts['Ansible']:
        generate_playbook(opts)
    else:
        load_policy(opts)

    print('\nRestart the container with: "--security-opt label=type:' + opts['ContainerName'] + '.process" parameter')

if __name__ == "__main__":
    main()
